@@ -19,12 +19,14 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 from papyon.msnp2p.constants import ApplicationID
+from papyon.msnp2p.errors import TLPParseError
 from papyon.util import debug
 from papyon.util.decorator import rw_property
 
 import struct
 import random
 import logging
+import uuid
 
 logger = logging.getLogger('papyon.msnp2p.transport')
 
@@ -77,8 +79,14 @@ class TLPHeader(object):
                 self.dw2,
                 self.qw1)
 
-    def parse(self, data):
-        fields = struct.unpack("<LLQQLLLLQ", data[:48])
+    def parse(self, data, chunk_size):
+        header = debug.hexify_string(data[:48])
+
+        try:
+            fields = struct.unpack("<LLQQLLLLQ", data[:48])
+        except:
+            raise TLPParseError(1, "invalid header", header)
+
         self.session_id = fields[0]
         self.blob_id = fields[1]
         self.blob_offset = fields[2]
@@ -89,6 +97,15 @@ class TLPHeader(object):
         self.dw2 = fields[7]
         self.qw1 = fields[8]
 
+        # if chunk is containing all blob data, chunk_size field might be 0
+        if self.blob_size == chunk_size and self.chunk_size == 0:
+            self.chunk_size = chunk_size
+
+        if self.blob_offset + self.chunk_size > self.blob_size:
+            raise TLPParseError(1, "chunk end exceeds blob size", header)
+        if self.blob_size >= 0 and self.chunk_size == 0:
+            raise TLPParseError(1, "empty chunk for non-empty blob", header)
+
 
 class MessageChunk(object):
     def __init__(self, header=TLPHeader(), body="", application_id=0):
@@ -153,6 +170,9 @@ class MessageChunk(object):
     def is_nonce_chunk(self):
         return self.header.flags & TLPFlag.KEY
 
+    def is_syn_request(self):
+        return False
+
     def is_data_preparation_chunk(self):
         return (self.header.chunk_size == 4 and self.header.blob_size == 4 and
                 self.body == "\x00\x00\x00\x00" and
@@ -181,7 +201,7 @@ class MessageChunk(object):
             return True
         return False
 
-    def create_ack_chunk(self):
+    def create_ack_chunk(self, sync=False):
         flags = TLPFlag.ACK
         if self.header.flags & TLPFlag.RAK:
             flags |= TLPFlag.RAK
@@ -204,19 +224,13 @@ class MessageChunk(object):
         bytes += struct.pack(">H", self.header.dw2 >> 16)
         bytes += struct.pack("<Q", self.header.qw1)
 
-        nonce = [("%X" % ord(byte)).zfill(2) for byte in bytes]
-        for idx in (4, 7, 10, 13):
-            nonce.insert(idx, '-')
-        return "".join(nonce)
+        return uuid.UUID(bytes=bytes)
 
     def set_nonce(self, nonce):
         """Set the chunk headers from a nonce and make it a nonce chunk by
            adding the KEY flag."""
 
-        nonce = filter(lambda c: c not in '{-}', nonce)
-        bytes = ""
-        for i in range(0, len(nonce), 2):
-            bytes += chr(int(nonce[i:i+2], 16))
+        bytes = nonce.bytes
 
         self.header.dw1 = struct.unpack(">L", bytes[0:4])[0]
         self.header.dw2 = struct.unpack(">H", bytes[4:6])[0]
@@ -224,7 +238,6 @@ class MessageChunk(object):
         self.header.qw1 = struct.unpack("<Q", bytes[8:16])[0]
         self.header.flags |= TLPFlag.KEY
 
-
     @staticmethod
     def create(app_id, session_id, blob_id, offset, blob_size, max_size, sync):
         header = TLPHeader()
@@ -237,8 +250,11 @@ class MessageChunk(object):
 
     @staticmethod
     def parse(data):
+        if len(data) < 48:
+            raise TLPParseError(1, "chunk should be at least 48 bytes")
+
         header = TLPHeader()
-        header.parse(data[:48])
+        header.parse(data[:48], len(data) - 48)
         body = data[48:]
         return MessageChunk(header, body)
 
