@@ -37,19 +37,19 @@ class BaseP2PTransport(gobject.GObject):
     __gsignals__ = {
             "chunk-received": (gobject.SIGNAL_RUN_FIRST,
                 gobject.TYPE_NONE,
-                (object,)),
+                (object, object, object)), # peer, peer_guid, chunk
 
             "chunk-sent": (gobject.SIGNAL_RUN_FIRST,
                 gobject.TYPE_NONE,
-                (object,)),
+                (object, object, object)), # peer, peer_guid, chunk
 
             "blob-received": (gobject.SIGNAL_RUN_FIRST,
                 gobject.TYPE_NONE,
-                (object,)),
+                (object, object, object)), # peer, peer_guid, blob
 
             "blob-sent": (gobject.SIGNAL_RUN_FIRST,
                 gobject.TYPE_NONE,
-                (object,)),
+                (object, object, object)), # peer, peer_guid, blob
             }
 
     def __init__(self, transport_manager, name):
@@ -62,9 +62,9 @@ class BaseP2PTransport(gobject.GObject):
         self._local_chunk_id = None
         self._remote_chunk_id = None
 
-        self._transport_manager._register_transport(self)
         self._queue_lock = threading.Lock()
         self._reset()
+        self._transport_manager._register_transport(self)
 
     @property
     def name(self):
@@ -93,24 +93,26 @@ class BaseP2PTransport(gobject.GObject):
     def can_send(self, peer, peer_guid, blob, bootstrap=False):
         raise NotImplementedError
 
+    def has_data_to_send(self):
+        return (len(self._data_blob_queue) > 0)
+
     def send(self, peer, peer_guid, blob):
         self._queue_lock.acquire()
-        self._data_blob_queue.append((peer, peer_guid, blob))
+        queue = self._data_blob_queue.setdefault(blob.session_id, [])
+        queue.append((peer, peer_guid, blob))
         self._queue_lock.release()
         self._start_processing()
 
     def cleanup(self, session_id):
         # remove this session's blobs from the data queue
         self._queue_lock.acquire()
-        canceled_blobs = []
-        for blob in self._data_blob_queue:
-            if blob[2].session_id == session_id:
-                canceled_blobs.append(blob)
-        for blob in canceled_blobs:
-            self._data_blob_queue.remove(blob)
+        if session_id in self._data_blob_queue:
+            del self._data_blob_queue[session_id]
         self._queue_lock.release()
 
     def close(self):
+        self._stop_processing()
+        self._reset()
         self._transport_manager._unregister_transport(self)
 
     def _ready_to_send(self):
@@ -124,10 +126,9 @@ class BaseP2PTransport(gobject.GObject):
     def _reset(self):
         self._queue_lock.acquire()
         self._first = True
-        self._data_blob_queue = []
-        self._pending_blob = {} # ack_id : (blob, callback, errback)
+        self._data_blob_queue = {} # session_id : [(peer, peer_guid, blob)]
+        self._outgoing_chunks = {} # chunk : blob
         self._pending_ack = set()
-        self._signaling_blobs = {} # blob_id : blob
         self._queue_lock.release()
 
     def _add_pending_ack(self, ack_id):
@@ -136,57 +137,34 @@ class BaseP2PTransport(gobject.GObject):
     def _del_pending_ack(self, ack_id):
         self._pending_ack.discard(ack_id)
 
-    def _add_pending_blob(self, ack_id, blob):
-        if self.version == 1:
-            self._pending_blob[ack_id] = blob
-        else:
-            self.emit("blob-sent", blob)
-
-    def _del_pending_blob(self, ack_id):
-        if not ack_id in self._pending_blob:
-            return
-        blob = self._pending_blob.pop(ack_id)
-        self.emit("blob-sent", blob)
-
     def _on_chunk_received(self, peer, peer_guid, chunk):
         if chunk.is_data_preparation_chunk():
             return
 
+        if self._first and not chunk.is_syn_request():
+            self._first = False
+
         if chunk.require_ack():
-            ack_chunk = chunk.create_ack_chunk()
+            ack_chunk = chunk.create_ack_chunk(self._first)
+            self._first = False
             self.__send_chunk(peer, peer_guid, ack_chunk)
 
         if chunk.is_ack_chunk() or chunk.is_nak_chunk():
             self._del_pending_ack(chunk.acked_id)
-            self._del_pending_blob(chunk.acked_id)
 
         #FIXME: handle all the other flags (NAK...)
 
         if not chunk.is_control_chunk():
-            if chunk.is_signaling_chunk(): # signaling chunk
-                self._on_signaling_chunk_received(chunk)
-            else: # data chunk (buffered by the transport manager)
-                self.emit("chunk-received", chunk)
+            self.emit("chunk-received", peer, peer_guid, chunk)
 
         self._start_processing()
 
-    def _on_signaling_chunk_received(self, chunk):
-        blob_id = chunk.blob_id
-        if blob_id in self._signaling_blobs:
-            blob = self._signaling_blobs[blob_id]
-        else:
-            # create an in-memory blob
-            blob = MessageBlob(chunk.application_id, "",
-                chunk.blob_size, chunk.session_id, blob_id)
-            self._signaling_blobs[blob_id] = blob
-
-        blob.append_chunk(chunk)
-        if blob.is_complete():
-            self.emit("blob-received", blob)
-            del self._signaling_blobs[blob_id]
-
-    def _on_chunk_sent(self, chunk):
-        self.emit("chunk-sent", chunk)
+    def _on_chunk_sent(self, peer, peer_guid, chunk):
+        self.emit("chunk-sent", peer, peer_guid, chunk)
+        blob = self._outgoing_chunks.pop(chunk, None)
+        if blob and blob.is_complete() and blob not in self._outgoing_chunks.values():
+            if not chunk.is_data_preparation_chunk():
+                self.emit("blob-sent", peer, peer_guid, blob)
         self._start_processing()
 
     def _start_processing(self):
@@ -202,25 +180,41 @@ class BaseP2PTransport(gobject.GObject):
     def _process_send_queue(self):
         if not self._queue_lock.acquire(False):
             return True
-        if len(self._data_blob_queue) == 0:
+        if not self.has_data_to_send():
             self._queue_lock.release()
-            self._stop_processing()
             return False
-        if not self._ready_to_send():
+
+        # FIXME find a better algorithm to choose session
+        if 0 in self._data_blob_queue:
+            session_id = 0
+        else:
+            session_id = self._data_blob_queue.keys()[0]
+
+        if session_id != 0 and not self._ready_to_send():
             logger.info("Transport is not ready to send, bail out")
             self._queue_lock.release()
-            self._stop_processing()
-            return True
+            return False
 
         sync = self._first
         self._first = False
-        (peer, peer_guid, blob) = self._data_blob_queue[0]
-        chunk = blob.get_chunk(self.version, self.max_chunk_size, sync)
+        (peer, peer_guid, blob) = self._data_blob_queue[session_id][0]
+
+        try:
+            chunk = blob.get_chunk(self.version, self.max_chunk_size, sync)
+        except Exception, err:
+            logger.exception(err)
+            logger.warning("Couldn't get chunk for session %s" % session_id)
+            self._data_blob_queue[session_id].pop(0) #ignoring blob
+            self._queue_lock.release()
+            return True
+
+        self._outgoing_chunks[chunk] = blob
         self.__send_chunk(peer, peer_guid, chunk)
 
         if blob.is_complete():
-            self._data_blob_queue.pop(0)
-            self._add_pending_blob(chunk.ack_id, blob)
+            self._data_blob_queue[session_id].pop(0)
+        if len(self._data_blob_queue[session_id]) == 0:
+            del self._data_blob_queue[session_id]
         self._queue_lock.release()
         return True
 
