@@ -18,10 +18,11 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-from papyon.msnp.message import MessageAcknowledgement
+from papyon.msnp.message import Message, MessageAcknowledgement
 from papyon.msnp2p.transport.TLP import MessageChunk
 from papyon.msnp2p.transport.base import BaseP2PTransport
 from papyon.switchboard_manager import SwitchboardHandler
+from papyon.util.parsing import build_account, parse_account
 
 import gobject
 import struct
@@ -36,11 +37,11 @@ class SwitchboardP2PTransport(BaseP2PTransport, SwitchboardHandler):
 
     MAX_OUTSTANDING_SENDS = 5
 
-    def __init__(self, client, switchboard, contacts, peer, peer_guid, transport_manager):
+    def __init__(self, client, contacts, peer, peer_guid, transport_manager):
         self._oustanding_sends = 0
         self._peer = peer
         self._peer_guid = peer_guid
-        SwitchboardHandler.__init__(self, client, switchboard, contacts)
+        SwitchboardHandler.__init__(self, client, contacts)
         BaseP2PTransport.__init__(self, transport_manager, "switchboard")
 
     def close(self):
@@ -48,27 +49,23 @@ class SwitchboardP2PTransport(BaseP2PTransport, SwitchboardHandler):
         self._leave()
 
     @staticmethod
-    def _can_handle_message(message, switchboard_client=None):
+    def _can_handle_message(message, self=None):
+        if self and (self.peer != message.sender or
+                self.peer_guid != message.sender_guid):
+            return False
         content_type = message.content_type[0]
         return content_type == 'application/x-msnmsgrp2p'
 
     @staticmethod
     def handle_peer(client, peer, peer_guid, transport_manager):
-        return SwitchboardP2PTransport(client, None, (peer,), peer, peer_guid,
+        return SwitchboardP2PTransport(client, (peer,), peer, peer_guid,
             transport_manager)
 
     @staticmethod
-    def handle_message(client, switchboard, message, transport_manager):
-        guid = None
-        peer = None
-        if 'P2P-Src' in message.headers and ';' in message.headers['P2P-Src']:
-            account, guid = message.headers['P2P-Src'].split(';', 1)
-            guid = guid[1:-1]
-            if account == client.profile.account:
-                peer = client.profile
-        if peer is None:
-            peer = switchboard.participants.values()[0]
-        return SwitchboardP2PTransport(client, switchboard, (), peer, guid,
+    def handle_message(client, message, transport_manager):
+        peer = message.sender
+        guid = message.sender_guid
+        return SwitchboardP2PTransport(client, (), peer, guid,
             transport_manager)
 
     @property
@@ -90,51 +87,61 @@ class SwitchboardP2PTransport(BaseP2PTransport, SwitchboardHandler):
     def can_send(self, peer, peer_guid, blob, bootstrap=False):
         return (self._peer == peer and self._peer_guid == peer_guid)
 
-    def __parse_guid(self, message, header):
-        if header not in message.headers or ';' not in message.headers[header]:
-            return None
-        return message.headers[header].split(';', 1)[1][1:-1]
-
     def _ready_to_send(self):
         return (self._oustanding_sends < self.MAX_OUTSTANDING_SENDS)
 
     def _send_chunk(self, peer, peer_guid, chunk):
         logger.debug(">>> %s" % repr(chunk))
+
+        msg = Message(self._client.profile)
         if chunk.version is 1 or peer_guid is None:
-            headers = {'P2P-Dest': self.peer.account}
+            msg.add_header('P2P-Dest', peer.account)
         elif chunk.version is 2:
-            headers = {'P2P-Src' : self._client.profile.account + ";{" +
-                                   self._client.machine_guid + "}",
-                       'P2P-Dest': peer.account + ";{" +
-                                   peer_guid + "}"}
-        content_type = 'application/x-msnmsgrp2p'
-        body = str(chunk) + struct.pack('>L', chunk.application_id)
+            msg.add_header('P2P-Src', build_account(self._client.profile.account,
+                                                    self._client.machine_guid))
+            msg.add_header('P2P-Dest', build_account(peer.account, peer_guid))
+        msg.content_type = 'application/x-msnmsgrp2p'
+        msg.body = str(chunk) + struct.pack('>L', chunk.application_id)
+
         self._oustanding_sends += 1
-        self._send_message(content_type, body, headers,
-                MessageAcknowledgement.MSNC, (self._on_message_sent, chunk),
-                (self._on_message_error, chunk))
+        self._send_message(msg, MessageAcknowledgement.MSNC,
+                (self._on_message_sent, peer, peer_guid, chunk),
+                (self._on_message_error, peer, peer_guid, chunk))
 
     def _on_message_received(self, message):
         version = 1
+        peer = message.sender
+        peer_guid = message.sender_guid
+
+        try:
+            dest, dest_guid = parse_account(message.get_header('P2P-Dest'))
+        except Exception, err:
+            logger.warning("Couldn't parse destination GUID: %s" % err)
+            return
+
         # if destination contains a GUID, the protocol should be TLPv2
-        dest_guid = self.__parse_guid(message, 'P2P-Dest')
-        src_guid = self.__parse_guid(message, 'P2P-Src')
-        if dest_guid and src_guid:
+        if dest_guid and peer_guid:
             version = 2
             if dest_guid != self._client.machine_guid or \
-               src_guid != self._peer_guid:
-                return # this chunk is not for us
+               peer_guid != self._peer_guid:
+                logger.debug("Received chunk is not for this end point")
+                return
+
+        try:
+            chunk = MessageChunk.parse(version, message.body[:-4])
+            chunk.application_id = struct.unpack('>L', message.body[-4:])[0]
+        except Exception, err:
+            logger.warning("Invalid TLP chunk in SB message: %s" % err)
+            return
 
-        chunk = MessageChunk.parse(version, message.body[:-4])
-        chunk.application_id = struct.unpack('>L', message.body[-4:])[0]
         logger.debug("<<< %s" % repr(chunk))
-        self._on_chunk_received(self._peer, self._peer_guid, chunk)
+        self._on_chunk_received(peer, peer_guid, chunk)
 
-    def _on_message_sent(self, chunk):
+    def _on_message_sent(self, peer, peer_guid, chunk):
         self._oustanding_sends -= 1
-        self._on_chunk_sent(chunk)
+        self._on_chunk_sent(peer, peer_guid, chunk)
 
-    def _on_message_error(self, chunk):
+    def _on_message_error(self, error, peer, peer_guid, chunk):
         self._oustanding_sends -= 1
 
     def _on_switchboard_closed(self):
@@ -144,7 +151,8 @@ class SwitchboardP2PTransport(BaseP2PTransport, SwitchboardHandler):
         BaseP2PTransport.close(self)
 
     def _on_error(self, error_type, error):
-        logger.info("Received error %i (type=%i)" % (error, error_type))
+        logger.info("Received error: %s (type=%i)" % (error, error_type))
+        self.close()
 
     def _on_contact_joined(self, contact):
         pass
