@@ -31,6 +31,7 @@ from papyon.profile import ContactType
 from papyon.service.AddressBook.constants import *
 from papyon.service.description.AB.constants import *
 from papyon.service.AddressBook.scenario.contacts import *
+from papyon.util.async import run
 
 import gobject
 
@@ -198,7 +199,7 @@ class AddressBook(gobject.GObject):
         """The address book object."""
         gobject.GObject.__init__(self)
 
-        self._ab = ab.AB(sso, proxies)
+        self._ab = ab.AB(sso, client, proxies)
         self._sharing = sharing.Sharing(sso, proxies)
         self._client = client
 
@@ -232,13 +233,10 @@ class AddressBook(gobject.GObject):
         self._state = AddressBookState.SYNCHRONIZING
 
         def callback(address_book, memberships):
-            ab = address_book.ab
-            contacts = address_book.contacts
-            groups = address_book.groups
-            for group in groups:
+            for group in address_book.groups:
                 g = profile.Group(group.Id, group.Name.encode("utf-8"))
                 self.groups.add(g)
-            for contact in contacts:
+            for contact in address_book.contacts:
                 c = self.__build_contact(contact, Membership.FORWARD)
                 if c is None:
                     continue
@@ -251,22 +249,28 @@ class AddressBook(gobject.GObject):
 
         initial_sync = scenario.InitialSyncScenario(self._ab, self._sharing,
                 (callback,),
-                (self.__common_errback, None),
+                (self.__common_errback,),
                 self._client.profile.account)
         initial_sync()
 
     # Public API
     def search_contact(self, account, network_id):
+        if account.lower() == self._client.profile.account.lower() and \
+                network_id == NetworkID.MSN:
+            return self._client.profile
+
         contacts = self.contacts.search_by_network_id(network_id).\
                 search_by_account(account)
         if len(contacts) == 0:
             return None
         return contacts[0]
 
-    def search_or_build_contact(self, account, network_id):
+    def search_or_build_contact(self, account, network_id, display_name=None):
         contact = self.search_contact(account, network_id)
         if contact is None:
-            contact = profile.Contact(None, network_id, account, account)
+            if not display_name:
+                display_name = account
+            contact = profile.Contact(None, network_id, account, display_name)
         return contact
 
     def check_pending_invitations(self, done_cb=None, failed_cb=None):
@@ -320,6 +324,9 @@ class AddressBook(gobject.GObject):
         contact = self.search_contact(account, network_id)
         old_memberships = (contact and contact.memberships) or Membership.NONE
 
+        if contact is self._client.profile:
+            return # can't add ourself to the address book
+
         if contact is not None and contact.is_mail_contact():
             self.upgrade_mail_contact(contact, groups, done_cb, failed_cb)
         elif contact is None or not contact.is_member(Membership.FORWARD):
@@ -599,6 +606,9 @@ class AddressBook(gobject.GObject):
             else:
                 continue
 
+            if network == NetworkID.MSN and member.IsPassportNameHidden:
+                continue # ignore contacts with hidden passport name
+
             contact = self.search_contact(member.Account, network)
             new_contact = False
             if contact is None:
@@ -607,10 +617,12 @@ class AddressBook(gobject.GObject):
                 account = member.Account.encode("utf-8")
                 display_name = (member.DisplayName or member.Account).encode("utf-8")
                 msg = member.Annotations.get('MSN.IM.InviteMessage', u'')
-                c = profile.Contact(None, network, account, display_name, cid)
-                c._server_attribute_changed('invite_message', msg.encode("utf-8"))
-                self.contacts.add(c)
-                contact = c
+                contact = profile.Contact(None, network, account, display_name, cid)
+                contact._server_attribute_changed('invite_message', msg.encode("utf-8"))
+                self.contacts.add(contact)
+
+            if contact is self._client.profile:
+                continue # don't update our own memberships
 
             for role in member.Roles:
                 membership = role_to_membership.get(role, None)
@@ -625,19 +637,11 @@ class AddressBook(gobject.GObject):
     def __common_callback(self, signal, callback, *args):
         if signal is not None:
             self.emit(signal, *args)
-        if callback is not None:
-            cb_args = args + callback[1:]
-            callback[0](*cb_args)
+        run(callback, *args)
 
-    def __common_errback(self, error_code, *args):
-        callback = args[-1]
-        args = args[:-1]
-        if callback is not None:
-            cb_args = args + callback[1:]
-            callback[0](error_code, *cb_args)
-        if error_code == AddressBookError.UNKNOWN:
-            # known errors are not fatal, just ignore them
-            self.emit('error', error_code)
+    def __common_errback(self, error, errback=None):
+        run(errback, error)
+        self.emit('error', error)
 
 gobject.type_register(AddressBook)
 
